import dash
import dash_bootstrap_components as dbc
from dash import html
from dash.development.base_component import Component

from amora.dash.components import question_details
from amora.models import list_models
from amora.questions import QUESTIONS

dash.register_page(
    __name__, title="Data Questions", fa_icon="fa-circle-question", location="sidebar"
)


def layout() -> Component:
    list(list_models())
    return html.Div(
        id="questions-content",
        children=dbc.CardGroup(
            children=[question_details.component(question) for question in QUESTIONS]
        ),
    )



from typing import NamedTuple

import dash
import dash_bootstrap_components as dbc
from dash import html
from dash.development.base_component import Component

SIDEBAR_STYLE = {
    "position": "fixed",
    "top": 0,
    "left": 0,
    "bottom": 0,
    "width": "24rem",
    "padding": "2rem 1rem",
}


class NavItem(NamedTuple):
    fa_icon: str
    href: str
    title: str


def nav() -> dbc.Nav:
    return dbc.Nav(
        [
            dbc.NavLink(
                [
                    html.I(className=f"fa-regular {page.get('fa_icon')}"),
                    " ",
                    page["name"],
                ],
                href=page["relative_path"],
                active="exact",
            )
            for page in dash.page_registry.values()
            if page.get("location") == "sidebar"
        ],
        vertical=True,
        pills=True,
    )


def component() -> Component:
    return html.Div(
        [
            html.H2("🌱 Amora", className="display-4"),
            html.Hr(),
            nav(),
        ],
        style=SIDEBAR_STYLE,
        id="side-bar",
    )



styles = {
    "container": {
        "position": "fixed",
        "display": "flex",
        "flex-direction": "column",
        "height": "100%",
        "width": "100%",
        "overflow": "scroll",
    },
}

theme = {
    "dark": True,
    "detail": "#007439",
    "primary": "#00EA64",
    "secondary": "#6E6E6E",
}



from typing import List

import dash
import dash_bootstrap_components as dbc
from dash import dcc, html
from dash.development.base_component import Component

from amora.dash.components import model_details
from amora.models import Model, list_models

dash.register_page(__name__, fa_icon="fa-database", location="sidebar")


def models_list(models: List[Model]) -> Component:
    options = [model.unique_name for model in models]

    return dcc.Dropdown(
        options=options,
        id="model-select-dropdown",
        value=options[0],
        placeholder="Select a model",
    )


def layout() -> Component:
    models = [model for (model, _path) in list_models()]

    return html.Div(
        id="models-content",
        children=[
            dbc.Row(
                [
                    dbc.Col(
                        [
                            models_list(models),
                            html.Div(
                                model_details.component(models[0]),
                                id="model-details",
                            ),
                        ]
                    ),
                ],
                align="start",
            )
        ],
    )



import pandas as pd
from sqlalchemy import Float, Integer, Numeric, func, literal

from amora.feature_store.protocols import FeatureViewSourceProtocol
from amora.models import Column, Model, select
from amora.providers.bigquery import run


def summarize(model: Model) -> pd.DataFrame:
    return pd.concat(
        [summarize_column(model, column) for column in model.__table__.columns]
    )


def summarize_column(model: Model, column: Column) -> pd.DataFrame:
    is_numeric = isinstance(column.type, (Numeric, Integer, Float))

    stmt = select(
        func.min(column).label("min"),
        func.max(column).label("max"),
        func.count(column.distinct()).label("unique_count"),
        (func.avg(column) if is_numeric else literal(None)).label("avg"),  # type: ignore
        (func.stddev(column) if is_numeric else literal(None)).label("stddev"),  # type: ignore
        ((literal(100) * func.countif(column == None)) / func.count(column)).label(
            "null_percentage"
        ),
    )
    result = run(stmt)

    df = pd.DataFrame.from_dict({k: [v] for k, v in dict(result.rows.next()).items()})
    df["column_name"] = column.name
    df["column_type"] = str(column.type)

    if isinstance(model, FeatureViewSourceProtocol):
        df["is_fv_feature"] = column.name in (
            c.name for c in model.feature_view_features()
        )
        df["is_fv_entity"] = column.name in (
            c.name for c in model.feature_view_entities()
        )
        df["is_fv_event_timestamp"] = (
            column.name == model.feature_view_event_timestamp().name
        )
    return df



"""
- Lista de feature views
    - data lineage
    - colunas
        - nome
        - tipo
    - nome
    - total de chaves na online store
    - summary da offline store
    - estimativa de custo da online store?
    - timestamp da última materialização

- cachear informações
"""
from typing import Iterable

import dash
import dash_bootstrap_components as dbc
from dash import html
from dash.development.base_component import Component
from feast import Feature, FeatureService, FeatureView

from amora.dag import DependencyDAG
from amora.dash.components import dependency_dag, model_summary
from amora.feature_store.registry import FEATURE_REGISTRY
from amora.meta_queries import summarize
from amora.models import Model, list_models

dash.register_page(__name__, fa_icon="fa-shopping-cart", location="sidebar")


def entities_list_items(entities: Iterable[str]):
    for entity in entities:
        yield dbc.ListGroupItem(entity, color="primary")


def features_list_items(features: Iterable[Feature]):
    for feature in features:
        yield dbc.ListGroupItem(feature.name)


def feature_details(fv: FeatureView, fs: FeatureService, model: Model) -> Component:
    summary = summarize(model)

    return dbc.Card(
        dbc.CardBody(
            [
                html.H5(fv.name, className="feature-view-name"),
                html.Small(
                    f"Latest time up to which the feature view has been materialized: '{fv.most_recent_end_time}'",
                    className="card-text text-muted",
                ),
                html.Div(
                    [
                        dbc.Row(model_summary.component(summary)),
                        dbc.Row(
                            [
                                dependency_dag.component(
                                    dag=DependencyDAG.from_model(model)
                                ),
                                dbc.Col("Bla"),
                            ]
                        ),
                    ]
                ),
            ]
        )
    )


def layout() -> Component:
    list(list_models())

    card_group = dbc.CardGroup(
        id="feature-store-card-group",
        children=[
            feature_details(fv, fs, model)
            for (fv, fs, model) in FEATURE_REGISTRY.values()
        ],
    )
    return html.Div(
        id="feature-store-content",
        children=[
            html.H1("Feature Store"),
            html.H2("Registered in this project are:"),
            card_group,
        ],
    )



from amora.config import settings
from amora.models import MetaData, create_engine

local_engine = create_engine(
    f"sqlite:///{settings.LOCAL_ENGINE_SQLITE_FILE_PATH}",
    echo=settings.LOCAL_ENGINE_ECHO,
)
local_metadata = MetaData(schema=None)



import dash
from dash import Dash, Input, Output, dcc, html
from dash.development.base_component import Component

from amora.dash.authentication import add_auth0_login
from amora.dash.components import model_details, side_bar
from amora.dash.config import settings
from amora.dash.css_styles import styles
from amora.models import amora_model_for_name

dash_app = Dash(
    __name__, external_stylesheets=settings.external_stylesheets, use_pages=True
)

if settings.auth0_login_enabled:
    add_auth0_login(dash_app)

# App
dash_app.layout = html.Div(
    style=styles["container"],
    children=[
        html.Div(
            [
                dcc.Location(id="url"),
                side_bar.component(),
                html.Div(
                    dash.page_container,
                    style={
                        "margin-left": "24rem",
                        "margin-right": "2rem",
                        "padding": "2rem 1rem",
                        "overflow": "scroll",
                    },
                    id="page-content",
                ),
            ],
        ),
    ],
)


@dash_app.callback(
    Output("model-details", "children"),
    Input("model-select-dropdown", "value"),
    prevent_initial_call=True,
)
def update_model_details(value: str) -> Component:
    return model_details.component(model=amora_model_for_name(value))



from typing import Dict, Iterable, List

import networkx as nx
from matplotlib import pyplot as plt

from amora.config import settings
from amora.materialization import Task
from amora.models import Model
from amora.utils import list_target_files


class DependencyDAG(nx.DiGraph):
    def __iter__(self):
        # todo: validar se podemos substituir por graphlib
        return nx.topological_sort(self)

    @classmethod
    def from_model(cls, model: Model) -> "DependencyDAG":
        """
        Builds the DependencyDAG for a given model
        """
        dag = cls()
        dag.add_node(model.unique_name)

        # fixme: percorrer de forma não recursiva
        def fetch_edges(node: Model):
            for dependency in getattr(node, "__depends_on__", []):
                dag.add_edge(dependency.unique_name, node.unique_name)
                fetch_edges(dependency)

        fetch_edges(model)
        return dag

    @classmethod
    def from_tasks(cls, tasks: Iterable[Task]) -> "DependencyDAG":
        dag = cls()

        for task in tasks:
            dag.add_node(task.model.unique_name)
            for dependency in getattr(task.model, "__depends_on__", []):
                dag.add_edge(dependency.unique_name, task.model.unique_name)

        return dag

    @classmethod
    def from_target(cls, models=None) -> "DependencyDAG":
        """
        Builds a DependencyDAG from the files compiled at `settings.AMORA_TARGET_PATH`

        :param models:
        :return:
        """
        model_to_task = {}

        for target_file_path in list_target_files():
            if models and target_file_path.stem not in models:
                continue

            task = Task.for_target(target_file_path)
            model_to_task[task.model.unique_name] = task

        return cls.from_tasks(tasks=model_to_task.values())

    def to_cytoscape_elements(self) -> List[Dict]:
        """

        Returns itself as a cytoscape schema compatible representation. E.g:

        For a `A --> B` graph:

        ```python
        [
            {"data": {"id": "A", "label": "A"}},
            {"data": {"id": "B", "label": "B"}},
            {"data": {"source": "A", "target": "B"}},
        ]
        ```
        """
        # todo: Estilizar nó de acordo com o tipo de materialização
        # todo: Adicionar metadados para cytoscape
        return [
            *({"data": {"id": node, "label": node}} for node in self.nodes),
            *(
                {"data": {"source": source, "target": target}}
                for source, target in self.edges
            ),
        ]

    def draw(self) -> None:
        plt.figure(1, figsize=settings.CLI_MATERIALIZATION_DAG_FIGURE_SIZE)
        nx.draw(
            self,
            with_labels=True,
            font_weight="bold",
            font_size="12",
            linewidths=4,
            node_size=150,
            node_color="white",
            font_color="green",
        )
        plt.show()

    def root(self):
        sorted_elements = list(self)
        if not sorted_elements:
            return None
        return sorted_elements[0]



import dash_cytoscape
from dash import html
from dash.development.base_component import Component

from amora.dag import DependencyDAG


def component(dag: DependencyDAG) -> Component:

    return html.Div(
        className="cy-container",
        # style=styles["cy-container"],
        children=[
            dash_cytoscape.Cytoscape(
                id="cytoscape-layout",
                elements=dag.to_cytoscape_elements(),
                layout={
                    "name": "breadthfirst",
                    "roots": f'[id = "{dag.root()}"]',
                    "refresh": 20,
                    "fit": True,
                    "padding": 30,
                    "randomize": False,
                },
                stylesheet=[
                    {"selector": "node", "style": {"label": "data(label)"}},
                    {
                        "selector": "edge",
                        "style": {
                            "curve-style": "bezier",
                            "target-arrow-shape": "triangle",
                        },
                    },
                ],
                responsive=True,
            )
        ],
    )



import dash_bootstrap_components as dbc
from dash import html
from dash.development.base_component import Component

from amora.dag import DependencyDAG
from amora.dash.components import dependency_dag, materialization_badge, model_summary
from amora.meta_queries import summarize
from amora.models import Model


def component(model: Model) -> Component:
    model_config = model.__model_config__
    return dbc.Card(
        [
            dependency_dag.component(DependencyDAG.from_model(model)),
            dbc.CardBody(
                [
                    html.H4(model.unique_name, className="card-title"),
                    materialization_badge.component(model_config.materialized),
                    html.P(
                        model_config.description,
                        className="card-text",
                    ),
                    html.Div([model_summary.component(summary=summarize(model))]),
                ]
            ),
        ],
    )



import dash
from dash import html
from dash.development.base_component import Component

from amora.config import settings
from amora.dag import DependencyDAG
from amora.dash.components import dependency_dag

dash.register_page(__name__, path="/", fa_icon="fa-house", location="sidebar")


def layout() -> Component:
    return html.Div(
        [
            html.H1(f"Project: {settings.TARGET_PROJECT}", id="title"),
            dependency_dag.component(dag=DependencyDAG.from_target()),
        ]
    )



import inspect
import re
from dataclasses import dataclass, field
from enum import Enum, auto
from importlib.util import module_from_spec, spec_from_file_location
from inspect import getfile
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Tuple, Type, Union

from sqlalchemy import Column, MetaData, Table, select
from sqlalchemy.orm import declared_attr
from sqlalchemy.sql import ColumnElement
from sqlmodel import Field, Session, SQLModel, create_engine

from amora.config import settings
from amora.logger import logger
from amora.protocols import CompilableProtocol
from amora.types import Compilable
from amora.utils import list_files, model_path_for_target_path

select = select
Column = Column
ColumnElement = ColumnElement
Columns = Iterable[ColumnElement]
Field = Field
Model = Type["AmoraModel"]
MetaData = MetaData
Models = Iterable[Model]
Session = Session
create_engine = create_engine


@dataclass
class PartitionConfig:
    field: str
    data_type: str = "date"
    granularity: str = "day"
    range: Optional[Dict[str, Any]] = None


class AutoName(Enum):
    def _generate_next_value_(name, start, count, last_values):
        return name


class MaterializationTypes(AutoName):
    ephemeral = auto()
    view = auto()
    table = auto()


@dataclass
class ModelConfig:
    """
    Model configuration metadata

    Attributes:
        cluster_by (List[str]): BigQuery tables can be [clustered](https://cloud.google.com/bigquery/docs/clustered-tables) to colocate related data. Expects a list of columns, as strings.
        description (Optional[str]): A string description of the model, used for documentation
        labels (Dict[str,str]): A dict of labels that can be used for resource selection
        materialized (amora.models.MaterializationTypes): The materialization configuration: `view`, `table`, `ephemeral`. Default: `view`
        partition_by (Optional[PartitionConfig]): BigQuery supports the use of a [partition by](https://cloud.google.com/bigquery/docs/partitioned-tables) clause to easily partition a table by a column or expression. This option can help decrease latency and cost when querying large tables.
    """

    description: str = "Undocumented! Generated by Amora Data Build Tool 💚"
    materialized: MaterializationTypes = MaterializationTypes.view
    partition_by: Optional[PartitionConfig] = None
    cluster_by: List[str] = field(default_factory=list)
    labels: Dict[str, str] = field(default_factory=dict)


metadata = MetaData(schema=f"{settings.TARGET_PROJECT}.{settings.TARGET_SCHEMA}")


class AmoraModel(SQLModel):
    """
    Attributes:
        __depends_on__ (Models): A list of Amora Models that the current model depends on
        __model_config__ (ModelConfig): Model configuration metadata
        __table__ (Table): SQLAlchemy table object
        __table_args__ (Dict[str, Any]): SQLAlchemy table arguments
    """

    __depends_on__: Models = []
    __model_config__ = ModelConfig(materialized=MaterializationTypes.view)
    __table__: Table
    __table_args__: Dict[str, Any] = {"extend_existing": True}
    metadata = metadata

    @declared_attr  # type: ignore
    def __tablename__(cls) -> str:  # type: ignore
        """
        By default, `__tablename__` is the `snake_case` class name.

        ```python
        class MyModel(AmoraModel):
            ...


        assert MyModel.__tablename__ == "my_model"
        ```
        """
        return re.sub(r"(?<!^)(?=[A-Z])", "_", cls.__name__).lower()

    @classmethod
    def dependencies(cls) -> Models:
        source = cls.source()
        if source is None:
            return []

        # todo: Remover necessidade de __depends_on__ inspecionando a query e chegando ao modelo de origem
        # tables: List[Table] = source.froms

        return cls.__depends_on__

    @classmethod
    def source(cls) -> Optional[Compilable]:
        """
        Called when `amora compile` is executed, Amora will build this model
        in your data warehouse by wrapping it in a `create view as` or `create table as` statement.

        Return `None` for defining models for tables/views that already exist on the data warehouse
        and shouldn't be managed by Amora.

        Return a `Compilable`, which is a sqlalchemy select statement, in order to compile the model with the given statement
        :return:
        """
        return None

    @classmethod
    def target_path(cls, model_file_path: Union[str, Path]) -> Path:
        # {settings.dbt_models_path}/a_model/a_model.py -> a_model/a_model.py
        strip_path = settings.MODELS_PATH.as_posix()
        relative_model_path = str(model_file_path).split(strip_path)[1][1:]
        # a_model/a_model.py -> ~/project/amora/target/a_model/a_model.sql
        target_file_path = settings.TARGET_PATH.joinpath(
            relative_model_path.replace(".py", ".sql")
        )

        return target_file_path

    @classmethod
    def model_file_path(cls) -> Path:
        return Path(getfile(cls))

    @classmethod
    @property
    def unique_name(cls) -> str:
        return str(cls.__table__)


def amora_model_for_path(path: Path) -> Model:
    spec = spec_from_file_location(".".join(["amoramodel", path.stem]), path)
    if spec is None:
        raise ValueError(f"Invalid path `{path}`. Not a valid Python file.")

    module = module_from_spec(spec)

    if spec.loader is None:
        raise ValueError(f"Invalid path `{path}`. Unable to load module.")

    try:
        spec.loader.exec_module(module)  # type: ignore
    except ImportError as e:
        raise ValueError(f"Invalid path `{path}`. Unable to load module.") from e
    is_amora_model = (
        lambda x: isinstance(x, CompilableProtocol)
        and inspect.isclass(x)
        and issubclass(x, AmoraModel)  # type: ignore
    )
    compilables = inspect.getmembers(
        module,
        is_amora_model,  # type: ignore
    )

    for _name, class_ in compilables:
        try:
            # fixme: Quando carregamos o código em inspect, não existe um arquivo associado,
            #  ou seja, ao iterar sobre as classes de um arquivo, a classe que retornar um TypeError,
            #  é uma classe definida no arquivo
            class_.model_file_path()
        except TypeError:
            return class_

    raise ValueError(f"Invalid path `{path}`")


def amora_model_for_target_path(path: Path) -> Model:
    model_path = model_path_for_target_path(path)
    return amora_model_for_path(model_path)


def amora_model_for_name(model_name: str) -> Model:
    for model, path in list_models():
        if model.unique_name == model_name:
            return model
    raise ValueError(f"{model_name} not found on models list")


def list_models(
    path: Path = settings.MODELS_PATH,
) -> Iterable[Tuple[Model, Path]]:
    for model_file_path in list_files(path, suffix=".py"):
        try:
            yield amora_model_for_path(model_file_path), model_file_path
        except ValueError:
            logger.warning(
                "Unable to load amora model for path",
                extra={"model_file_path": model_file_path},
            )
            continue



from dataclasses import dataclass
from datetime import date, datetime, time
from enum import Enum
from typing import Any, Dict, Iterable, List, Optional, Union

import sqlalchemy
from google.api_core.client_info import ClientInfo
from google.cloud.bigquery import (
    Client,
    QueryJobConfig,
    SchemaField,
    Table,
    TableReference,
)
from google.cloud.bigquery.table import RowIterator, _EmptyRowIterator
from sqlalchemy import (
    ARRAY,
    JSON,
    TIMESTAMP,
    Boolean,
    Date,
    DateTime,
    Float,
    Integer,
    String,
    Time,
    func,
    literal,
    literal_column,
)
from sqlalchemy.sql import coercions, expression, operators, roles, sqltypes
from sqlalchemy.sql.selectable import CTE
from sqlalchemy_bigquery.base import BQArray, unnest
from sqlmodel import AutoString

from amora.compilation import compile_statement
from amora.config import settings
from amora.contracts import BaseResult
from amora.models import Column, ColumnElement, Model, select
from amora.types import Compilable
from amora.version import VERSION

Schema = List[SchemaField]
BQTable = Union[Table, TableReference, str]

# todo: cobrir todos os tipos
BIGQUERY_TYPES_TO_PYTHON_TYPES = {
    "ARRAY": list,
    "BIGNUMERIC": int,
    "BOOL": bool,
    "BOOLEAN": bool,
    "BYTES": bytes,
    "DATE": date,
    "DATETIME": datetime,
    "FLOAT64": float,
    "FLOAT": float,
    "GEOGRAPHY": str,
    "INT64": int,
    "INTEGER": int,
    "JSON": dict,
    "STRING": str,
    "TIME": time,
    "TIMESTAMP": datetime,
}

SQLALCHEMY_TYPES_TO_BIGQUERY_TYPES = {
    Integer: "INTEGER",
    String: "STRING",
    AutoString: "STRING",
    DateTime: "DATETIME",
    Date: "DATE",
    Time: "TIME",
    Float: "FLOAT",
    Boolean: "BOOLEAN",
    JSON: "JSON",
    TIMESTAMP: "TIMESTAMP",
}


class TimePart(Enum):
    """
    Since BigQuery represents time parts as an unquoted value,
    we need to use `literal_column`

    More on: https://cloud.google.com/bigquery/docs/reference/standard-sql/time_functions#time_trunc
    """

    MICROSECOND = literal_column("MICROSECOND")
    MILLISECOND = literal_column("MILLISECOND")
    SECOND = literal_column("SECOND")
    MINUTE = literal_column("MINUTE")
    HOUR = literal_column("HOUR")


@dataclass
class DryRunResult(BaseResult):
    model: Model
    schema: Schema

    @property
    def estimated_cost(self):
        return estimated_query_cost_in_usd(self.total_bytes)


@dataclass
class RunResult(BaseResult):
    rows: Union[RowIterator, _EmptyRowIterator]
    execution_time_in_ms: int
    schema: Optional[Schema] = None

    @property
    def estimated_cost(self):
        return estimated_query_cost_in_usd(self.total_bytes)


_client = None


def get_client() -> Client:
    global _client
    if _client is None:
        _client = Client(
            client_info=ClientInfo(
                client_library_version=VERSION,
                user_agent=f"amora-data-build-tool/{VERSION}",
            )
        )
    return _client


def get_fully_qualified_id(model: Model) -> str:
    return f"{model.metadata.schema}.{model.__tablename__}"


def get_schema(table_id: str) -> Schema:
    """
    Given a `table_id`, returns the `Schema` of the table by querying BigQueries API
    """
    client = get_client()
    table = client.get_table(table_id)
    return table.schema


def get_schema_for_model(model: Model) -> Schema:
    """
    Given an `AmoraModel`, returns the equivalent bigquery `Schema`
    of the model by parsing the model SQLAlchemy column schema
    """
    columns = model.__table__.columns

    def gen_schema():
        for col in columns:
            is_array = isinstance(col.type, ARRAY)
            if is_array:
                mode = "REPEATED"
                field_type = SQLALCHEMY_TYPES_TO_BIGQUERY_TYPES[
                    col.type.item_type.__class__
                ]
            else:
                mode = "NULLABLE"
                field_type = SQLALCHEMY_TYPES_TO_BIGQUERY_TYPES[col.type.__class__]

            yield SchemaField(
                name=col.name,
                field_type=field_type,
                mode=mode,
            )

    return list(gen_schema())


def get_schema_for_source(model: Model) -> Optional[Schema]:
    """
    Give an `Amora Model`, returns the bigquery `Schema` of its
    `source` classmethod query result
    """
    source = model.source()
    if source is None:
        return None

    result = dry_run(model)
    if result is None:
        return None

    return result.schema


def run(statement: Compilable) -> RunResult:
    """
    Executes a given query and returns its results
    and metadata as an `amora.providers.bigquery.RunResult`
    """
    query = compile_statement(statement)
    query_job = get_client().query(query)
    rows = query_job.result()
    execution_time_delta = query_job.ended - query_job.started

    return RunResult(
        execution_time_in_ms=execution_time_delta.microseconds / 1000,
        job_id=query_job.job_id,
        query=query,
        referenced_tables=[
            ".".join(table.to_api_repr().values())
            for table in query_job.referenced_tables
        ],
        rows=rows,
        schema=query_job.schema,
        total_bytes=query_job.total_bytes_billed,
        user_email=query_job.user_email,
    )


def dry_run(model: Model) -> Optional[DryRunResult]:
    """
    You can use the estimate returned by the dry run to calculate query
    costs in the pricing calculator. Also useful to verify user permissions
    and query validity. You are not charged for performing the dry run.

    Read more: https://cloud.google.com/bigquery/docs/dry-run-queries

    E.g:
    ```python
    dry_run(HeartRate)
    ```

    Will result in:

    ```python
    DryRunResult(
        total_bytes_processed=170181834,
        query=\"SELECT\\n  `health`.`creationDate`,\\n  `health`.`device`,\\n  `health`.`endDate`,\\n  `health`.`id`,\\n  `health`.`sourceName`,\\n  `health`.`startDate`,\\n  `health`.`unit`,\\n  `health`.`value`\\nFROM `diogo`.`health`\\nWHERE `health`.`type` = 'HeartRate'\",
        model=HeartRate,
        referenced_tables=["amora-data-build-tool.diogo.health"],
        schema=[
            SchemaField("creationDate", "TIMESTAMP", "NULLABLE", None, (), None),
            SchemaField("device", "STRING", "NULLABLE", None, (), None),
            SchemaField("endDate", "TIMESTAMP", "NULLABLE", None, (), None),
            SchemaField("id", "INTEGER", "NULLABLE", None, (), None),
            SchemaField("sourceName", "STRING", "NULLABLE", None, (), None),
            SchemaField("startDate", "TIMESTAMP", "NULLABLE", None, (), None),
            SchemaField("unit", "STRING", "NULLABLE", None, (), None),
            SchemaField("value", "FLOAT", "NULLABLE", None, (), None),
        ],
    )
    ```
    """
    client = get_client()
    source = model.source()
    if source is None:
        table = client.get_table(get_fully_qualified_id(model))

        if table.table_type == "VIEW":
            query_job = client.query(
                query=table.view_query,
                job_config=QueryJobConfig(dry_run=True, use_query_cache=False),
            )
            return DryRunResult(
                job_id=query_job.job_id,
                model=model,
                query=table.view_query,
                referenced_tables=[
                    ".".join(table.to_api_repr().values())
                    for table in query_job.referenced_tables
                ],
                schema=query_job.schema,
                total_bytes=query_job.total_bytes_processed,
                user_email=query_job.user_email,
            )
        else:
            return DryRunResult(
                job_id=None,
                model=model,
                query=None,
                referenced_tables=[str(table.reference)],
                schema=table.schema,
                total_bytes=table.num_bytes,
                user_email=None,
            )

    query = compile_statement(source)

    query_job = client.query(
        query=query,
        job_config=QueryJobConfig(dry_run=True, use_query_cache=False),
    )
    return DryRunResult(
        job_id=query_job.job_id,
        total_bytes=query_job.total_bytes_processed,
        referenced_tables=[
            ".".join(table.to_api_repr().values())
            for table in query_job.referenced_tables
        ],
        query=query,
        model=model,
        schema=query_job.schema,
        user_email=query_job.user_email,
    )


class fixed_unnest(sqlalchemy.sql.roles.InElementRole, unnest):
    _with_offset = None

    def __init__(self, *args, **kwargs):
        self.name = "unnest"
        super().__init__(*args, **kwargs)

    def table_valued(self, *expr, with_offset: str = None, **kwargs):
        new_func = self._generate()

        if with_offset:
            expr += (with_offset,)
            new_func._with_offset = with_offset

        new_func.type = new_func._table_value_type = sqltypes.TableValueType(*expr)

        return new_func


def cte_from_rows(rows: Iterable[Dict[str, Any]]) -> CTE:
    """
    Returns a table like selectable (CTE) for the given hardcoded values.

    E.g:
    ```python
    rows = [{"numeric_column": "123"}, {"numeric_column": "234"}, {"numeric_column": "345"}]
    cte_from_rows(rows)
    ```

    Will result in the following SQL

    ```sql
        WITH `annon_1` AS (
            SELECT "123" AS numeric_column
            UNION ALL SELECT "234 AS numeric_column
            UNION ALL SELECT "345" AS numeric_column
        )
    ```

    Which would render a table like:

    ```md
    | numeric_column |
    |----------------|
    | 123            |
    | 234            |
    | 345            |
    ```

    Useful both for model writing and testing purposes. Think of `cte_from_rows` as way of generating
    a "temporary table like object", with data available at runtime.

    """

    def gen_selects(rows):
        for row in rows:
            cols = []
            for name, value in row.items():
                if isinstance(value, array):
                    cols.append(value.label(name))
                else:
                    cols.append(literal(value).label(name))
            yield select(cols)

    selects = list(gen_selects(rows))

    if len(selects) == 1:
        return selects[0].cte()
    else:
        return selects[0].union_all(*(selects[1:])).cte()


def estimated_query_cost_in_usd(total_bytes: int) -> float:
    """
    By default, queries are billed using the on-demand pricing model,
    where you pay for the data scanned by your queries.

    - This function doesn't take into consideration that the first 1 TB per month is free.
    - By default, the estimation is based on BigQuery's `On-demand analysis` pricing, which may change over time and
    may vary according to regions and your personal contract with GCP.

    You may set `AMORA_GCP_BIGQUERY_ON_DEMAND_COST_PER_TERABYTE_IN_USD` to the appropriate value for your use case.

    More on: https://cloud.google.com/bigquery/pricing#analysis_pricing_models

    :param total_bytes: Total data processed by the query
    :return: The estimated cost in USD, based on `On-demand` price
    """
    total_terabytes = total_bytes / 1024**4
    return total_terabytes * settings.GCP_BIGQUERY_ON_DEMAND_COST_PER_TERABYTE_IN_USD


def estimated_storage_cost_in_usd(total_bytes: int) -> float:
    """
    Storage pricing is the cost to store data that you load into BigQuery.
    `Active storage` includes any table or table partition that has been modified in the last 90 days.

    - This function doesn't take into consideration that the first 10 GB of storage per month is free.
    - By default, the estimation is based on BigQuery's `Active Storage` cost per GB, which may change over time and
    may vary according to regions and your personal contract with GCP.

    You may set `AMORA_GCP_BIGQUERY_ACTIVE_STORAGE_COST_PER_GIGABYTE_IN_USD` to the appropriate value for your use case.

    More on: https://cloud.google.com/bigquery/pricing#storage

    :param total_bytes: Total bytes stored into the table
    :return: The estimated cost in USD, based on `Active storage` price
    """
    total_gigabytes = total_bytes / 1024**3
    return (
        total_gigabytes * settings.GCP_BIGQUERY_ACTIVE_STORAGE_COST_PER_GIGABYTE_IN_USD
    )


class array(expression.ClauseList, expression.ColumnElement):  # type: ignore
    """
    A BigQuery ARRAY literal.

    This is used to produce `ARRAY` literals in SQL expressions, e.g.:

    ```python
    from sqlalchemy import select

    from amora.compilation import compile_statement
    from amora.providers.bigquery import array

    stmt = select([array([1, 2]).label("a"), array([3, 4, 5]).label("b")])

    compile_statement(stmt)
    ```
    Produces the SQL:

    ```sql
    SELECT
        ARRAY[1, 2] AS a,
        ARRAY[3, 4, 5]) AS b
    ```

    An instance of `array` will always have the datatype `sqlalchemy_bigquery.base.BQArray`.
    The "inner" type of the array is inferred from the values present, unless the
    ``type_`` keyword argument is passed, e.g.:

    ```python
    array(["foo", "bar"], type_=String)
    ```
    """

    __visit_name__ = "array"

    def __init__(self, clauses, **kw):
        clauses = [coercions.expect(roles.ExpressionElementRole, c) for c in clauses]
        super().__init__(*clauses, **kw)
        self._type_tuple = [arg.type for arg in clauses]
        main_type = kw.pop(
            "type_",
            self._type_tuple[0] if self._type_tuple else sqltypes.NULLTYPE,
        )
        self.type = BQArray(main_type, dimensions=1)

        for type_ in self._type_tuple:
            if type(type_) is sqltypes.NullType:
                raise ValueError("Array cannot have a null element")

    def _bind_param(self, operator, obj, _assume_scalar=False, type_=None):
        if _assume_scalar or operator is operators.getitem:
            # if getitem->slice were called, Indexable produces
            # a Slice object from that
            assert isinstance(obj, int)
            return expression.BindParameter(
                None,
                obj,
                _compared_to_operator=operator,
                type_=type_,
                _compared_to_type=self.type,
                unique=True,
            )

        else:
            return array(
                [
                    self._bind_param(operator, o, _assume_scalar=True, type_=type_)
                    for o in obj
                ]
            )

    def self_group(self, against=None):
        if against in (operators.any_op, operators.all_op, operators.getitem):
            return expression.Grouping(self)
        else:
            return self


def zip_arrays(
    *arr_columns: Column, additional_columns: Optional[List[Column]] = None
) -> Compilable:
    """
    Given at least two array columns of equal size, return a table of the unnest values,
    converting array items into rows. E.g:

    A CTE with 3 array columns: `entity`, `f1`, `f2`

    ```python
    from amora.providers.bigquery import array, cte_from_rows, zip_arrays

    cte = cte_from_rows(
        [
            {
                "entity": array([1, 2]),
                "f1": array(["f1v1", "f1v2"]),
                "f2": array(["f2v1", "f2v2"]),
            }
        ]
    )

    zip_arrays(cte.c.entity, cte.c.f1, cte.c.f2)
    ```

    Will result in the following table:

    | entity | f1   | f2   |
    |--------|------|------|
    | 1      | f1v1 | f2v1 |
    | 2      | f1v2 | f2v2 |

    If additional columns are needed from the original data, those can be selected
    using the optional `additional_columns`:

    ```python
    from amora.providers.bigquery import array, cte_from_rows, zip_arrays

    cte = cte_from_rows(
        [
            {
                "entity": array([1, 2]),
                "f1": array(["f1v1", "f1v2"]),
                "f2": array(["f2v1", "f2v2"]),
                "id": 1,
            },
            {
                "entity": array([3, 4]),
                "f1": array(["f1v3", "f1v4"]),
                "f2": array(["f2v3", "f2v4"]),
                "id": 2,
            },
        ]
    )

    zip_arrays(cte.c.entity, cte.c.f1, cte.c.f2, additional_columns=[cte.c.id])
    ```

    | entity | f1   | f2   | id   |
    |--------|------|------|------|
    | 1      | f1v1 | f2v1 | 1    |
    | 2      | f1v2 | f2v2 | 1    |
    | 3      | f1v3 | f2v3 | 2    |
    | 4      | f1v4 | f2v4 | 2    |

    Read more: [https://cloud.google.com/bigquery/docs/reference/standard-sql/arrays#zipping_arrays](https://cloud.google.com/bigquery/docs/reference/standard-sql/arrays#zipping_arrays)
    """
    offset_alias = "off"
    offset = func.offset(literal_column(offset_alias))

    columns: List[ColumnElement] = [col[offset].label(col.key) for col in arr_columns]
    if additional_columns:
        columns += additional_columns

    return select(columns).join(
        fixed_unnest(arr_columns[0]).table_valued(with_offset=offset_alias),
        onclause=literal(1) == literal(1),
        isouter=True,
    )



import logging
import sys

from amora.config import settings

logger = logging.getLogger("amora")
logger.setLevel(settings.LOGGER_LOG_LEVEL)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))



import dash_bootstrap_components as dbc
from dash import dash_table, html
from dash.development.base_component import Component

from amora.questions import Question
from amora.visualization import Visualization, VisualizationKind


def answer_visualization(visualization: Visualization) -> Component:
    if visualization.config.kind == VisualizationKind.big_number:
        big_number = visualization.data.get_value()
        return html.P()
    elif visualization.config.kind == VisualizationKind.table:
        return dash_table.DataTable(
            columns=[
                {"name": col, "id": col, "selectable": True}
                for col in visualization.data.columns.values
            ],
            data=visualization.data.to_dict("records"),
            row_selectable="multi",
            sort_action="native",
            style_cell={
                "overflow": "hidden",
                "textOverflow": "ellipsis",
                "maxWidth": 0,
            },
            export_format="csv",
            export_headers="display",
        )


def component(question: Question) -> Component:
    return dbc.Card(
        className="question-card",
        children=dbc.CardBody(
            [
                html.H5(question.name, className="card-title"),
                answer_visualization(question.render()),
                dbc.Accordion(
                    [
                        dbc.AccordionItem(
                            html.Code(question.sql, lang="sql"), title="SQL"
                        )
                    ],
                    start_collapsed=True,
                ),
            ]
        ),
    )



import logging
import os
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Tuple
from uuid import uuid4

from pydantic import BaseSettings

ROOT_PATH = Path(__file__).parent.parent
AMORA_MODULE_PATH = ROOT_PATH.joinpath("amora")

_Width = float
_Height = float


class Settings(BaseSettings):
    TARGET_PROJECT: str
    TARGET_SCHEMA: str
    TARGET_PATH: Path
    MODELS_PATH: Path

    CLI_CONSOLE_MAX_WIDTH: int = 160
    CLI_MATERIALIZATION_DAG_FIGURE_SIZE: Tuple[_Width, _Height] = (32, 32)

    # https://cloud.google.com/bigquery/pricing#analysis_pricing_models
    GCP_BIGQUERY_ON_DEMAND_COST_PER_TERABYTE_IN_USD: float = 5.0
    # https://cloud.google.com/bigquery/pricing#storage
    GCP_BIGQUERY_ACTIVE_STORAGE_COST_PER_GIGABYTE_IN_USD: float = 0.020

    LOCAL_ENGINE_ECHO: bool = False
    LOCAL_ENGINE_SQLITE_FILE_PATH: Path = Path(
        NamedTemporaryFile(suffix="amora-sqlite.db", delete=False).name
    )
    LOGGER_LOG_LEVEL: int = logging.DEBUG

    MONEY_DECIMAL_PLACES: int = 4

    TEST_RUN_ID: str = os.getenv("PYTEST_XDIST_TESTRUNUID") or f"amora-{uuid4().hex}"

    class Config:
        env_prefix = "AMORA_"


settings = Settings()



from enum import Enum
from typing import NamedTuple, Union

import pandas as pd


class VisualizationKind(str, Enum):
    big_number = "big_number"
    bar_chart = "bar_chart"
    column_chart = "column_chart"
    line_chart = "line_chart"
    table = "table"
    pie_chart = "pie_chart"


class VisualizationConfig(NamedTuple):
    title: Union[str, None]
    kind: VisualizationKind = VisualizationKind.table


class Visualization:
    """
    The Amora visual representation of a `pandas.DataFrame`
    """

    def __init__(self, data: pd.DataFrame, config: VisualizationConfig):
        self.data = data
        self.config = config

    def __str__(self):
        return self.data.to_markdown()
